package quimufu.colourful_portals.client;

import me.jellysquid.mods.sodium.client.model.quad.ModelQuad;
import me.jellysquid.mods.sodium.client.model.quad.ModelQuadView;
import me.jellysquid.mods.sodium.client.model.quad.ModelQuadViewMutable;
import me.jellysquid.mods.sodium.client.model.quad.properties.ModelQuadFacing;
import me.jellysquid.mods.sodium.client.model.quad.properties.ModelQuadWinding;
import me.jellysquid.mods.sodium.client.render.chunk.compile.buffers.ChunkModelBuilder;
import me.jellysquid.mods.sodium.client.render.vertex.type.ChunkVertexEncoder;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.client.texture.Sprite;
import net.minecraft.fluid.FluidState;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Direction;
import net.minecraft.world.BlockRenderView;

@Environment(value = EnvType.CLIENT)
public class SodiumPortalFluidRenderHandler implements CommonPortalFluidRenderer.VertexEater {

    private Sprite sprite = null;
    private final ModelQuadViewMutable quad = new ModelQuad();

    private final CommonPortalFluidRenderer commonPortalFluidRenderer = new CommonPortalFluidRenderer();
    private final ChunkVertexEncoder.Vertex[] vertices = ChunkVertexEncoder.Vertex.uninitializedQuad();
    private int i;
    private ChunkModelBuilder chunkModelBuilder;

    public SodiumPortalFluidRenderHandler() {
    }

    public boolean render(BlockRenderView world, FluidState fluidState, BlockPos pos, BlockPos offset, ChunkModelBuilder buffers) {
        this.chunkModelBuilder = buffers;
        return commonPortalFluidRenderer.render(world,fluidState,pos,offset,this);
    }


    @Override
    public void setSprite(Sprite sprite) {
        quad.setSprite(sprite);
    }

    @Override
    public void eatVertex(float x, float y, float z, float u, float v) {
        quad.setX(i, x);
        quad.setY(i, y);
        quad.setZ(i, z);
        quad.setTexU(i, u);
        quad.setTexV(i, v);
        i = (i+1)%4;

    }

    @Override
    public void drawQuad(BlockPos offset, Direction direction) {
        writeQuad(chunkModelBuilder, offset, quad, ModelQuadFacing.fromDirection(direction), ModelQuadWinding.CLOCKWISE);
    }

    private void writeQuad(ChunkModelBuilder builder, BlockPos offset, ModelQuadView quad, ModelQuadFacing facing, ModelQuadWinding winding) {
        var vertexBuffer = builder.getVertexBuffer();
        var vertices = this.vertices;

        for (int i = 0; i < 4; i++) {
            var out = vertices[i];
            out.x = offset.getX() + quad.getX(i);
            out.y = offset.getY() + quad.getY(i);
            out.z = offset.getZ() + quad.getZ(i);
            out.color = 0xFFFFFFFF;
            out.u = quad.getTexU(i);
            out.v = quad.getTexV(i);
            out.light = 16;
        }

        Sprite sprite = quad.getSprite();

        if (sprite != null) {
            builder.addSprite(sprite);
        }

        builder.getIndexBuffer(facing)
                .add(vertexBuffer.push(vertices), winding);
    }



}
