package quimufu.colourful_portals.portal;

import net.minecraft.block.Block;
import net.minecraft.block.BlockState;
import net.minecraft.registry.Registries;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.Identifier;
import net.minecraft.util.math.BlockBox;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Direction;
import net.minecraft.util.math.Vec3i;

import java.util.*;

import static quimufu.colourful_portals.ColourfulPortalsMod.PORTAL_BLOCK;
import static quimufu.colourful_portals.ColourfulPortalsMod.PORTAL_FLUID;

public class PortalHelper {
    public static final char[][] PORTAL = new char[][]{
            {'N', 'X', 'X', 'N'},
            {'X', 'O', 'O', 'X'},
            {'X', 'O', 'O', 'X'},
            {'X', 'O', 'O', 'X'},
            {'N', 'X', 'X', 'N'},
    };

    static Iterator<BlockPos> insideOf(BlockBox portal) {
        Direction axisW = getAxisW(portal);
        Vec3i dir = axisW.getVector();
        BlockPos startPos = new BlockPos(portal.getMinX(), portal.getMinY(), portal.getMinZ());

        return new BlockPosIterator(startPos, dir, 'O');

    }

    static boolean isValidPortal(ServerWorld world, BlockBox portal, Identifier blockId) {
        return isValidPortal(world, portal, blockId, true, false);
    }

    static boolean isValidCandidate(ServerWorld world, BlockBox portal, Identifier blockId) {
        return isValidPortal(world, portal, blockId, false, false);
    }

    private static boolean isValidPortal(ServerWorld world, BlockBox portal, Identifier blockId, boolean empty, boolean placementCheck) {
        Direction axisW = getAxisW(portal);
        return isValidPortal(world, Registries.BLOCK.get(blockId), new BlockPos(portal.getMinX(), portal.getMinY(), portal.getMinZ()), empty, axisW, placementCheck);
    }

    static Direction getAxisW(BlockBox fromPortalBox) {
        if (fromPortalBox.getMaxX() - fromPortalBox.getMinX() > 2)
            return Direction.EAST;
        return Direction.SOUTH;
    }

    static List<BlockBox> findPortalCandidates(ServerWorld world, BlockPos pos, Identifier blockId) {
        Block block = Registries.BLOCK.get(blockId);
        ArrayList<BlockBox> portals = new ArrayList<>();
        outer:
        for (Direction direction : Direction.values()) {
            int size = direction == Direction.UP || direction == Direction.DOWN ? 5 : 4;
            //quick checks that fail on most non-portal buildings
            for (int i = 1; i < size - 1; i++) {
                if (world.getBlockState(pos.add(direction.getVector().multiply(i))).isOf(block)) {
                    continue outer;
                }
            }
            if (!world.getBlockState(pos.add(direction.getVector().multiply(size - 1))).isOf(block)) {
                continue;
            }
            //just check all possible portal shapes
            switch (direction) {
                case DOWN -> {
                    //orientation X wise
                    //start left
                    BlockPos lowerCorner = pos.add(-1, -4, 0);

                    getPortalStartingAt(world, block, lowerCorner, Direction.EAST)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(-2, -4, 0);

                    getPortalStartingAt(world, block, lowerCorner, Direction.EAST)
                            .ifPresent(portals::add);
                    //orientation Y wise
                    lowerCorner = pos.add(0, -4, -1);
                    getPortalStartingAt(world, block, lowerCorner, Direction.SOUTH)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(0, -4, -2);
                    getPortalStartingAt(world, block, lowerCorner, Direction.SOUTH)
                            .ifPresent(portals::add);
                }
                case UP -> {
                    //orientation X wise
                    //start left
                    BlockPos lowerCorner = pos.add(-1, 0, 0);

                    getPortalStartingAt(world, block, lowerCorner, Direction.EAST)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(-2, 0, 0);

                    getPortalStartingAt(world, block, lowerCorner, Direction.EAST)
                            .ifPresent(portals::add);
                    //orientation Y wise
                    lowerCorner = pos.add(0, 0, -1);
                    getPortalStartingAt(world, block, lowerCorner, Direction.SOUTH)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(0, 0, -2);
                    getPortalStartingAt(world, block, lowerCorner, Direction.SOUTH)
                            .ifPresent(portals::add);
                }
                case NORTH -> {
                    BlockPos lowerCorner = pos.add(0, -1, -3);
                    getPortalStartingAt(world, block, lowerCorner, Direction.SOUTH)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(0, -2, -3);
                    getPortalStartingAt(world, block, lowerCorner, Direction.SOUTH)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(0, -3, -3);
                    getPortalStartingAt(world, block, lowerCorner, Direction.SOUTH)
                            .ifPresent(portals::add);
                }
                case SOUTH -> {
                    BlockPos lowerCorner = pos.add(0, -1, 0);
                    getPortalStartingAt(world, block, lowerCorner, Direction.SOUTH)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(0, -2, 0);
                    getPortalStartingAt(world, block, lowerCorner, Direction.SOUTH)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(0, -3, 0);
                    getPortalStartingAt(world, block, lowerCorner, Direction.SOUTH)
                            .ifPresent(portals::add);
                }
                case WEST -> {
                    BlockPos lowerCorner = pos.add(-3, -1, 0);

                    getPortalStartingAt(world, block, lowerCorner, Direction.EAST)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(-3, -2, 0);

                    getPortalStartingAt(world, block, lowerCorner, Direction.EAST)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(-3, -3, 0);

                    getPortalStartingAt(world, block, lowerCorner, Direction.EAST)
                            .ifPresent(portals::add);
                }
                case EAST -> {
                    BlockPos lowerCorner = pos.add(0, -1, 0);

                    getPortalStartingAt(world, block, lowerCorner, Direction.EAST)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(0, -2, 0);

                    getPortalStartingAt(world, block, lowerCorner, Direction.EAST)
                            .ifPresent(portals::add);
                    lowerCorner = pos.add(0, -3, 0);

                    getPortalStartingAt(world, block, lowerCorner, Direction.EAST)
                            .ifPresent(portals::add);
                }
            }

        }
        return portals;
    }


    private static Optional<BlockBox> getPortalStartingAt(ServerWorld world, Block block, BlockPos lowerCorner, Direction direction) {
        Vec3i sidewardsOffset = direction.getVector().multiply(3);
        Vec3i upwardsOffset = Direction.UP.getVector().multiply(4);
        if (isValidPortal(world, block, lowerCorner, false, direction, false)) {
            return Optional.of(BlockBox.create(lowerCorner, lowerCorner.add(sidewardsOffset).add(upwardsOffset)));
        }
        return Optional.empty();
    }

    public static boolean isPortalPlaceable(ServerWorld world, BlockBox portal, Identifier blockId) {
        return isValidPortal(world, portal, blockId, false, true);
    }

    private static boolean isValidPortal(ServerWorld world, Block block, BlockPos lowerCorner, boolean hasToBeFilledCorrectly, Direction direction, boolean placementCheck) {
        //maybe use BlockPosIterator, but make sure Performance is OK
        for (int i = 0; i < 5; i++) {
            for (int j = 0; j < 4; j++) {
                Vec3i sidewardsOffset = direction.getVector().multiply(j);
                Vec3i upwardsOffset = Direction.UP.getVector().multiply(i);
                BlockPos curr = lowerCorner.add(sidewardsOffset).add(upwardsOffset);

                switch (PORTAL[i][j]) {
                    case 'X' -> {
                        if (!world.getBlockState(curr).isOf(block)) {
                            return false;
                        }
                    }
                    case 'N' -> {
                    }
                    case 'O' -> {
                        BlockState blockState = world.getBlockState(curr);
                        if (hasToBeFilledCorrectly && !(blockState.getFluidState().isOf(PORTAL_FLUID) || blockState.getBlock().equals(PORTAL_BLOCK))) {
                            return false;
                        }
                        if(placementCheck && !(blockState.isReplaceable() || blockState.getFluidState().isOf(PORTAL_FLUID) || blockState.getBlock().equals(PORTAL_BLOCK))){
                            return false;
                        }
                        if (blockState.isOf(block)) {
                            return false;
                        }
                    }

                }
            }
        }
        return true;
    }


    private static class BlockPosIterator implements Iterator<BlockPos> {
        private final BlockPos startPos;
        private final Vec3i dir;
        private final char filter;
        int pos;
        int size;
        static final Vec3i up = Direction.UP.getVector();

        public BlockPosIterator(BlockPos startPos, Vec3i dir, char filter) {
            this.startPos = startPos;
            this.dir = dir;
            this.filter = filter;
            this.size = PORTAL.length * PORTAL[0].length;
        }

        @Override
        public boolean hasNext() {
            return peekNext() != null;
        }

        @Override
        public BlockPos next() {
            while (pos < size) {
                int i = pos / PORTAL[0].length;
                int j = pos % PORTAL[0].length;
                pos++;
                if (PORTAL[i][j] == filter) {
                    return startPos.add(up.multiply(i)).add(dir.multiply(j));
                }
            }
            throw new NoSuchElementException();
        }

        private BlockPos peekNext() {
            int pos = this.pos;
            while (pos < size) {
                int i = pos / PORTAL[0].length;
                int j = pos % PORTAL[0].length;
                pos++;
                if (PORTAL[i][j] == filter) {
                    return startPos.add(up.multiply(i)).add(dir.multiply(j));
                }
            }
            return null;
        }

    }
}
